//******************************************************************
//******************************************************************
//** Author : Syed Salman Alam
//** Modified for use on FreeRTOS / Nordic NRF52
//** Custome code for Reading, Writing and erasing data to and from SD card
//** every function is tested. It works for both SD and SDHC cards
//******************************************************************
//******************************************************************
// Originally downloaded from https://devzone.nordicsemi.com/blogs/28/sd-card-interfacing/
// Numerous attempts have been performed to initialize the SD card interface, all resulting in:
// No bit pattern detected whatsoever on MISO during the 74 clocks, sending CMD0, clocking up to 255 bytes hoping for some low bit on MISO, loop ....
// The NRF52 supplies only 2.8V to the SD card. When using an external 3.3V power supply, and skipping CMD0 by sending CMD8, the answer appears to obscure the MOSI bit patterns as a bi-directional pin interface ??

#include "nrf.h"
#include "nrf_delay.h"
#include "boards.h"
#include <string.h>
#include <stdio.h>
#include "app_error.h"
#include "nrf_gpio.h"
#include "common.h"

#include "app_util_platform.h"
#include "nrf_drv_spi.h"
#include "nrf_drv_config.h"
#include "SD_Routines.h"

#define FAT_TESTING_ONLY

// Commands need to be bitwise-orred with 0x40.
// When bitwise orred with 0xC0, it is considered an application command and actually the sequence CMD55 + CMDxx is sent.
#define GO_IDLE_STATE            (0x40+0)
#define SEND_OP_COND             (0x40+1)
#define SEND_CSD                 (0x40+9)
#define STOP_TRANSMISSION        (0x40+12)
#define SEND_STATUS              (0xC0+13)
#define SET_BLOCK_LEN            (0x40+16)
#define READ_SINGLE_BLOCK        (0x40+17)
#define READ_MULTIPLE_BLOCKS     (0x40+18)
#define WRITE_SINGLE_BLOCK       (0x40+24)
#define WRITE_MULTIPLE_BLOCKS    (0x40+25)
#define ERASE_BLOCK_START_ADDR   (0x40+32)
#define ERASE_BLOCK_END_ADDR     (0x40+33)
#define ERASE_SELECTED_BLOCKS    (0x40+38)
#define CRC_ON_OFF               (0x40+59)
// Additionals by BR, from http://elm-chan.org/docs/mmc/mmc_e.html
#define APP_SEND_OP_COND         (0xC0+41) // For only SDC. Initiate initialization process.
#define	SEND_IF_COND             (0x40+8)  // For only SDC V2. Check voltage range.
#define SEND_CID            	 (0x40+10) // Read CID register.
#define APP_CMD					 (0x40+55) // Leading command of ACMD<n> command, e.g. command 41
#define READ_OCR            	 (0x40+58) // Read OCR.

// Control the power to the SD card via an output PIN
#define MMCSD_PIN_POWER			 17 // Currently not connected but too much trouble to remove all code

// Because we're debugging the non-functional initialization from both hard- and software, DAT1 and DAT2 are also connected
#define MMCSD_PIN_DAT1			 15
#define MMCSD_PIN_DAT2			 16

// Chip select explicitly (not via SPI controller)
#define MMCSD_PIN_SELECT 		 11
#if (SPI0_ENABLED == 1)
	#define SPI_INSTANCE		 0
	#define MMCSD_PIN_MOSI		 SPI0_CONFIG_MOSI_PIN
	#define MMCSD_PIN_SCK		 SPI0_CONFIG_SCK_PIN
	#define MMCSD_PIN_MISO		 SPI0_CONFIG_MISO_PIN
#elif (SPI1_ENABLED == 1)
	#define SPI_INSTANCE		 1
	#define MMCSD_PIN_MOSI		 SPI1_CONFIG_MOSI_PIN
	#define MMCSD_PIN_SCK		 SPI1_CONFIG_SCK_PIN
	#define MMCSD_PIN_MISO		 SPI1_CONFIG_MISO_PIN
#elif (SPI2_ENABLED == 1)
	#define SPI_INSTANCE		 2
	#define MMCSD_PIN_MOSI		 SPI2_CONFIG_MOSI_PIN
	#define MMCSD_PIN_SCK		 SPI2_CONFIG_SCK_PIN
	#define MMCSD_PIN_MISO		 SPI2_CONFIG_MISO_PIN
#else
	#error No SPIx_ENABLED controller configured, see nrf_drv_config.h file
#endif

#define ON     1
#define OFF    0

static const nrf_drv_spi_t spi_master = NRF_DRV_SPI_INSTANCE(SPI_INSTANCE);

uint32_t totalBlocks = 0; // Size of SD card currently mounted
static uint8_t spi_initialized = 0; // Flag if SPI interface is already initialized
static uint8_t cmd_wait_reply = 10; // Number of bytes to wait for an answer after sending a command
static uint8_t sd_buffer[512]; // Read-write block data buffer

//#define uart_print(...) ;
#define uart_print(...) printf(__VA_ARGS__)


//******************************************************************
//******************************************************************
//  Local functionality
//******************************************************************
//******************************************************************


static void mmcsd_select()
{
    nrf_gpio_pin_clear(MMCSD_PIN_SELECT);
}

static void mmcsd_deselect()
{
    nrf_gpio_pin_set(MMCSD_PIN_SELECT);
}

static uint8_t spi_xfer(uint8_t data)
{
    uint8_t spi_txbuf[2];
    uint8_t spi_rxbuf[2];

    spi_txbuf[0]=data;
    nrf_drv_spi_transfer(&spi_master, spi_txbuf, 1, spi_rxbuf, 1);
    return spi_rxbuf[0];
}

//******************************************************************
//Function: to initialize SPI and  SD/SDHC  Cards
//Arguments: none
//Note: SPI_SD_PIN_init() should have been called already
//return: 1 if SD Card, 2 if SDHC card, 0 if error
//******************************************************************
static uint8_t SPI_SD_init(void)
{
    uint32_t err = NRF_SUCCESS;

    uart_print(__func__);

	totalBlocks = 0UL;
    nrf_gpio_pin_clear( MMCSD_PIN_POWER);
    mmcsd_deselect();

    nrf_drv_spi_config_t conf =
    {
        .ss_pin 		= NRF_DRV_SPI_PIN_NOT_USED,
        .irq_priority	= APP_IRQ_PRIORITY_LOW,
        .orc 			= 0xCC,
        .frequency 		= NRF_DRV_SPI_FREQ_250K, // 100-400 kHz according http://elm-chan.org/docs/mmc/mmc_e.html
        .mode 			= NRF_DRV_SPI_MODE_0,
        .bit_order 		= NRF_DRV_SPI_BIT_ORDER_MSB_FIRST,
        .sck_pin  		= MMCSD_PIN_SCK,
        .mosi_pin 		= MMCSD_PIN_MOSI,
        .miso_pin 		= MMCSD_PIN_MISO
    };

    err = nrf_drv_spi_init(&spi_master, &conf, NULL); // NULL means blocking I/O
    nrf_gpio_cfg_input(MMCSD_PIN_MISO, NRF_GPIO_PIN_PULLUP); // from: https://devzone.nordicsemi.com/question/61472/nrf51-spim-sd-card/

    APP_ERROR_CHECK(err);

    return err;
}

// Based on http://www.mars.dti.ne.jp/~m7030/pic_room/rec/mmc.c
//******************************************************************
//Function: to send a command to SD card
//Arguments: unsigned char (8-bit command value)
// & unsigned long (32-bit command argument)
//return: unsigned char; response byte
//******************************************************************
static uint8_t SD_sendCommand(uint8_t cmd, uint32_t arg)
{
    uint8_t response, retry = cmd_wait_reply;

	if (cmd & 0x80) {	/* ACMD<n> is the command sequence of CMD55-CMD<n> */
		response = SD_sendCommand(APP_CMD, 0x00000000); // one time recursion
		if (response > 1) return response;
	}
	else
	{
		spi_xfer(0xFF); // Allow one idle byte before sending the command byte
	}

	// Send command & argument
    spi_xfer((cmd & 0x3F) | 0x40); //send command, first two bits always '01'
    spi_xfer(arg>>24);
    spi_xfer(arg>>16);
    spi_xfer(arg>>8);
    spi_xfer(arg);

	// Send CRC as well
	switch(cmd)
	{
			case GO_IDLE_STATE: spi_xfer(0x95); break;
			case SEND_IF_COND:  spi_xfer(0x87); break;
			default: 			spi_xfer(0x01); break;
	}

	// Receive response, timeout in max 80 clocks / 10 bytes
    while((response = spi_xfer(0xFF)) & 0x80) // wait R1 response (highest bit should be zero: http://elm-chan.org/docs/mmc/mmc_e.html)
        if(retry-- == 0) break; 			  // time out error

	if ( (response != 0xFF) && (cmd_wait_reply > 1) )
	{
		// We seemed to have received an answer, clock it away until we're idle again
		retry = 8;
	    while(spi_xfer(0xFF) != 0xFF) // wait idle
	        if(retry-- == 0) break;   // time out error
	}

    return response;
}

//******************************************************************
//Function: to initialize the SDSC card 
//Arguments: none
//return: 0 if no error, 1 if error
//******************************************************************
static uint8_t SDSC_init(void)
{
    static uint8_t rcv = 0, i;
    static uint16_t count=0;

	totalBlocks = 0UL;

	// Idle CS
	mmcsd_deselect();
	for(i = 0; i < 10; i++)
		spi_xfer( 0xFF); // Generate at least 74 idle clocks before CS

	mmcsd_select();
	spi_xfer( 0xFF);     // 16 clocks idle after CS, from https://courses.cit.cornell.edu/ee476/FinalProjects/s2007/cd247_maw72/cd247_maw72/index_files/image007.jpg
	do
	{
		rcv = SD_sendCommand(		GO_IDLE_STATE, 0x00000000); // CMD0
		if (rcv == 0x01) break;
		mmcsd_deselect();
		count++;
		nrf_delay_ms(1);
		mmcsd_select();
	} while (count < 800);
	if (rcv == 0x01)
	{
		count = 0;
		do
		{
			rcv = SD_sendCommand(	SEND_IF_COND, 0x000001AA); // CMD8
			if (rcv == 0x01) break;
			count++;
			for(i = 0; i < 10; i++) // Force a small delay
				mmcsd_select();
		} while (count < 20);


		SD_sendCommand(				CRC_ON_OFF, 		OFF); // CMD59 disable CRC; default - CRC disabled in SPI mode
		SD_sendCommand(				SET_BLOCK_LEN, 		512); // CMD16 set block size to 512 bytes

		totalBlocks = SD_readNbrBlocks();

	    mmcsd_deselect();
	    spi_xfer( 0xFF);

		return 0;
	}

    mmcsd_deselect();
    spi_xfer( 0xFF);

	return 1;
}

//******************************************************************
//Function: to Initialize SDHC card
//Arguments: non
//return: 0 if no error, 1 if error
//******************************************************************
static uint8_t SDHC_init(void)
{
    uint8_t rcv = 0;
    uint16_t count = 0;

    mmcsd_select();
    mmcsd_select();
    mmcsd_select();
	SD_sendCommand(GO_IDLE_STATE, 0x00000000); // CMD0
	SD_sendCommand(SEND_IF_COND,  0x000001AA); // CMD8

    while (count < 5000)
    {
    	rcv = SD_sendCommand(APP_SEND_OP_COND, 0x40000000); // ACMD41
        if (rcv == 0x00 )
        {
        	rcv = SD_sendCommand(READ_OCR, 0x00000000);

        	if (rcv == 0xC0)
            {
                spi_xfer(0xFF);
                spi_xfer(0xFF);
                spi_xfer(0xFF);
                spi_xfer(0xFF);

				totalBlocks = SD_readNbrBlocks();

				mmcsd_deselect();
                return 0;
            }
			break;
        }
        count++;
    }

    mmcsd_deselect();
    spi_xfer( 0xFF);

    return 1;
}


//******************************************************************
//******************************************************************
//  Externally available functionality
//******************************************************************
//******************************************************************


//******************************************************************
//Function: to initialize (SPI) SD card pins
//Arguments: none
//******************************************************************
void SPI_SD_PIN_init(void)
{
    nrf_gpio_cfg_output(MMCSD_PIN_POWER);
    nrf_gpio_pin_clear( MMCSD_PIN_POWER); // Power off to start with
    nrf_gpio_cfg_output(MMCSD_PIN_SELECT);
    mmcsd_deselect();                     // Idle CS to start with

    // Set pull-up for all other logic signals
    nrf_gpio_cfg_input(MMCSD_PIN_MISO, NRF_GPIO_PIN_PULLUP); // from: https://devzone.nordicsemi.com/question/61472/nrf51-spim-sd-card/
    nrf_gpio_cfg_input(MMCSD_PIN_MOSI, NRF_GPIO_PIN_PULLUP); // will be corrected by SPI_SD_init()
    nrf_gpio_cfg_input(MMCSD_PIN_SCK,  NRF_GPIO_PIN_PULLUP); // will be corrected by SPI_SD_init()

    nrf_gpio_cfg_input(MMCSD_PIN_DAT1, NRF_GPIO_PIN_PULLUP); // debug, to make sure there's no bad influence
    nrf_gpio_cfg_input(MMCSD_PIN_DAT2, NRF_GPIO_PIN_PULLUP); // debug, to make sure there's no bad influence
}

//******************************************************************
//Function: Initialize SPI interface (when required) and SD card
//          when present, SC or HC type, whatever is detected.
//Arguments: none
//return: SD_CARD_NONE_ERROR if error or no card detected
//        SD_CARD_TYPE_SC for standard capacity (1, 2 or 4 GB) card detected
//        SD_CARD_TYPE_HC for high capacity card detected
//******************************************************************
SD_CARD_TYPE SD_init()
{
	SD_CARD_TYPE ret = SD_CARD_NONE_ERROR;
	int          err = NRF_SUCCESS;

	/* Initialize SPI interface if not yet done */
	if ( !spi_initialized ) {
		err = SPI_SD_init();
		spi_initialized = 1;
	}
	else
	{
		mmcsd_deselect();
	}

	if (err == NRF_SUCCESS)
	{
		nrf_gpio_pin_set(MMCSD_PIN_POWER);
		nrf_delay_ms(2); // 2 ms sleep between power-up and communications: from http://elm-chan.org/docs/mmc/mmc_e.html / http://elm-chan.org/docs/mmc/pic/sdinit.png

		/* Initialize SD card, start by checking for SC card */
		if (SDSC_init())
		{
			/* SC init failed, check if we have a HC card */
			if(SDHC_init())
					ret = SD_CARD_NONE_ERROR;	// No or not recognized card
			else
					ret = SD_CARD_TYPE_HC;		// SDHC card
		}
		else
				ret = SD_CARD_TYPE_SC;				// SDSC card
	}
	if (ret == SD_CARD_NONE_ERROR)
	{
		nrf_gpio_pin_clear(MMCSD_PIN_POWER);
	}

	return ret;
}


//******************************************************************
//Function: De-initialize SD card (when present)
//Arguments: none
//return: SD_CARD_NONE_ERROR (always)
//******************************************************************
SD_CARD_TYPE SD_close()
{
	int i;

	totalBlocks = 0UL;

	if (spi_initialized)
	{
		mmcsd_select();
		SD_sendCommand(STOP_TRANSMISSION, 0x00000000); // CMD12
		for(i = 0; i < 12; i++)
		{
			spi_xfer( 0xFF); // Generate at least 74 idle clocks after command
			if (i == 8)
				mmcsd_deselect(); // Idle CS and generate some more clocks still
		}
	}

	return SD_CARD_NONE_ERROR;
}


//*****************************************************************
//Function: to erase specified no. of blocks of SD card
//param: starting block address
//param: number of blocks to erase
//return: unsigned char; will be 0 if no error,
// otherwise the response byte will be sent
// problem it is wring FF when erasing sector
//*****************************************************************
uint8_t SD_erase (uint32_t startBlock, uint32_t totalBlocks)
{
    uint8_t response;

    response = SD_sendCommand(ERASE_BLOCK_START_ADDR, startBlock<<9); //send starting block address
    if(response != 0x00) //check for SD status: 0x00 - OK (No flags set)
        return response;

    response = SD_sendCommand(ERASE_BLOCK_END_ADDR,(startBlock + totalBlocks - 1)<<9); //send end block address
    if(response != 0x00)
        return response;

    response = SD_sendCommand(ERASE_SELECTED_BLOCKS, 0); //erase all selected blocks
    if(response != 0x00)
        return response;

    return 0; //normal return
}

//*****************************************************************
//Function: to erase specified no. of blocks of SD card
//param: starting block address
//param: number of blocks to erase
//return: unsigned char; will be 0 if no error,
// otherwise the response byte will be sent
// problem it is wring FF when erasing sector
//*****************************************************************
void SD_erase_Modified(uint32_t startBlock, uint32_t totalBlocks)
{

    uint8_t erase_with[512];
    uint32_t c;
    memset(erase_with,'\0',512);

    for(c = startBlock; c < totalBlocks; c++)
        SD_writeSingleBlock(c, erase_with);
}




//******************************************************************
//Function: to read a single block from SD card
//param: block address
//return: unsigned char; will be 0 if no error,
// otherwise the response byte will be sent
//******************************************************************
// BR Split send cmd & actual read
static uint8_t SD_read(uint8_t *p_buffer)
{
    uint16_t i, retry=0;

    mmcsd_select();

    while(spi_xfer( 0xFF) != 0xFE) //wait for start block token 0xfe (0x11111110)
        if(retry++ > 0xfffe){ mmcsd_deselect(); return 1; } //return if time-out

    for(i=0; i<512; i++) //read 512 bytes
        sd_buffer[i] = spi_xfer( 0xFF);

    spi_xfer( 0xFF); //receive incoming CRC (16-bit), CRC is ignored here
    spi_xfer( 0xFF);

    spi_xfer( 0xFF); //extra 8 clock pulses
    mmcsd_deselect();

    return 0;
}

uint8_t SD_readSingleBlock(uint32_t startBlock, uint8_t *p_buffer)
{
    uint8_t response;

    response = SD_sendCommand(READ_SINGLE_BLOCK, startBlock << 9); //read a Block command
    //block address converted to starting address of 512 byte Block
    if(response == 0x00) { //check for SD status: 0x00 - OK (No flags set)
				response = SD_read(p_buffer);
		}

		return response;
}


// Based on http://playground.arduino.cc/Code/SDCARD, function sdc_totalNbrBlocks()
// get nbr of blocks on SD memory card from
long SD_readNbrBlocks(void) {
    uint8_t response;

    response = SD_sendCommand(SEND_CSD, 0);
    if(response == 0x00) { //check for SD status: 0x00 - OK (No flags set)
				response = SD_read(sd_buffer);
				if(response == 0x00) { //check for SD status: 0x00 - OK (No flags set)
						// compute size
						long C_Size = ((sd_buffer[0x08] & 0xC0) >> 6) | ((sd_buffer[0x07] & 0xFF) << 2) | ((sd_buffer[0x06] & 0x03) << 10);
						long C_Mult = ((sd_buffer[0x08] & 0x80) >> 7) | ((sd_buffer[0x08] & 0x03) << 2);
						long size = ((C_Size+1) << (C_Mult+2));
						uart_print("Card size %ld / 0x%X", size, (unsigned int)size);
						return size; 
				}
				else uart_print("SD read CSD data fail..");
		}
		else uart_print("SD send CSD command fail..");

		return (0x80000000 / 512); // Assume 2 GB
}

//******************************************************************
//Function: to write to a single block of SD card
//param: block address
//param: pointer to data buffer(data buffer should be of size 512)
//return: unsigned char; will be 0 if no error,
// otherwise the response byte will be sent
//******************************************************************
uint8_t SD_writeSingleBlock(uint32_t startBlock, uint8_t *p_buffer)
{
    uint8_t response;
    uint16_t i, retry=0;

    response = SD_sendCommand(WRITE_SINGLE_BLOCK, startBlock<<9); //write a Block command
    if(response != 0x00) //check for SD status: 0x00 - OK (No flags set)
        return response;

    mmcsd_select();
    //spi_xfer( 0xFF)
    spi_xfer(0xfe);     //Send start block token 0xfe (0x11111110)

    for(i=0; i<512; i++)    //send 512 bytes data
        spi_xfer(p_buffer[i]);

    spi_xfer(0xff);     //transmit dummy CRC (16-bit), CRC is ignored here
    spi_xfer(0xff);

    response = spi_xfer( 0xFF);

    if( (response & 0x1f) != 0x05) //response= 0xXXX0AAA1 ; AAA='010' - data accepted
    {                              //AAA='101'-data rejected due to CRC error
        mmcsd_deselect();              //AAA='110'-data rejected due to write error
        return response;
    }

    while(!spi_xfer( 0xFF)) //wait for SD card to complete writing and get idle
        if(retry++ > 0xfffe){mmcsd_deselect(); return 1;}

    mmcsd_deselect();
    spi_xfer(0xFF);   //just spend 8 clock cycle delay before reasserting the CS line
    mmcsd_select();          //re-asserting the CS line to verify if card is still busy

    while(!spi_xfer( 0xFF)) //wait for SD card to complete writing and get idle
        if(retry++ > 0xfffe){mmcsd_deselect(); return 1;}
    mmcsd_deselect();

    return 0;
}


//***************************************************************************
//Function: to read multiple blocks from SD card & send every block to UART
//param: starting block address
//param: ending block address
//return: unsigned char; will be 0 if no error,
// otherwise the response byte will be sent
//****************************************************************************
void Read_Sectors(uint32_t startBlock, uint32_t totalBlocks)
{
    uint32_t c_bytes,blocks_2print;
    for(blocks_2print=startBlock; blocks_2print<totalBlocks; blocks_2print++) //read 512 bytes
    {
        memset(sd_buffer,'\0',sizeof(sd_buffer));
        SD_readSingleBlock(blocks_2print, sd_buffer);
        for(c_bytes=0; c_bytes<512; c_bytes++) //read 512 bytes
        {
            printf( "%c" ,sd_buffer[c_bytes]);
            // nrf_delay_ms(1);
        }
    }
}




